###############################################################################
#
# Library for ikos regression tests
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2011-2019 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import argparse
import atexit
import os
import shutil
import sqlite3
import subprocess
import sys
import tempfile

USE_COLORS = True
INTERACTIVE = True
VERBOSE = False
CLANG = 'clang'
IKOS_PP = 'ikos-pp'
IKOS_ANALYZER = 'ikos-analyzer'

# available ikos analyses
ANALYSES = (
    'boa',
    'dbz',
    'nullity',
    'prover',
    'upa',
    'uva',
    'sio',
    'uio',
    'shc',
    'poa',
    'pcmp',
    'sound',
    'fca',
    'dca',
    'dfa',
    'dbg',
    'watch',
)

# available colors
COLORS = {
    'grey': 0,
    'red': 1,
    'green': 2,
    'yellow': 3,
    'blue': 4,
    'magenta': 5,
    'cyan': 6,
    'white': 7
}
# available attributes
ATTRIBUTES = {
    'bold': 1,
    'dark': 2,
    'underline': 4,
    'blink': 5,
    'reverse': 7,
    'concealed': 8
}


def printf(fmt, *args, **kwargs):
    file = kwargs.pop('file', sys.stdout)
    file.write(fmt % args if args else fmt)
    file.flush()


def colorize(text, color=None, on_color=None, attrs=None):
    '''
    Colorize text.

    >>> colorize('Hello, World!', 'red', 'grey', ['bold', 'blink'])
    '\x1b[5m\x1b[1m\x1b[40m\x1b[31mHello, World!\x1b[0m'
    >>> colorize('Hello, World!', 'green')
    '\x1b[32mHello, World!\x1b[0m'
    '''
    if USE_COLORS:
        if color:
            text = '\033[%dm%s' % (30 + COLORS[color], text)

        if on_color:
            text = '\033[%dm%s' % (40 + COLORS[on_color], text)

        if attrs:
            for attr in attrs:
                text = '\033[%dm%s' % (ATTRIBUTES[attr], text)

        text += '\033[0m'

    return text


def bold(s): return colorize(s, attrs=['bold'])
def red(s): return colorize(s, 'red', attrs=['bold'])
def green(s): return colorize(s, 'green', attrs=['bold'])
def yellow(s): return colorize(s, 'yellow', attrs=['bold'])


def is_executable(fpath):
    return fpath and os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    ''' Try to find program in the PATH, otherwise return None.

    >>> which('cat')
    '/bin/cat'
    '''
    fpath, fname = os.path.split(program)
    if fpath:
        if is_executable(program):
            return program
    else:
        for path in os.environ['PATH'].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_executable(exe_file):
                return exe_file

    return None


def find_clang():
    path = which(CLANG)
    assert is_executable(path), 'could not find clang'
    return path


def find_ikos_pp():
    path = which(IKOS_PP)
    assert is_executable(path), 'could not find ikos-pp'
    return path


def find_ikos_analyzer():
    path = which(IKOS_ANALYZER)
    assert is_executable(path), 'could not find ikos-analyzer'
    return path


def clang_emit_llvm_flags():
    ''' Clang flags to emit llvm bitcode '''
    # see analyzer.clang_emit_llvm_flags()
    return ['-c', '-emit-llvm']


def clang_ikos_flags():
    ''' Clang flags for ikos '''
    # see analyzer.clang_ikos_flags()
    return [
        # enable clang warnings
        '-Wall',
        # disable source code fortification
        '-U_FORTIFY_SOURCE',
        '-D_FORTIFY_SOURCE=0',
        # flag for intrinsic.h
        '-D__IKOS__',
        # compile in debug mode
        '-g',
        # disable optimizations
        '-O0',
        # disable the 'optnone' attribute
        # see https://bugs.llvm.org/show_bug.cgi?id=35950#c10
        '-Xclang',
        '-disable-O0-optnone',
    ]


class Result:
    OK = 0
    WARNING = 1
    ERROR = 2
    UNREACHABLE = 3


class Database:
    def __init__(self, path):
        self.db = sqlite3.connect(path)
        self.cursor = self.db.cursor()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.db.close()

    def get_num_checks(self, result=None):
        where = ('WHERE status=%d' % result) if result else ''
        self.cursor.execute('SELECT COUNT(*) FROM checks %s' % where)
        return self.cursor.fetchone()[0]

    def get_line_status(self, line):
        self.cursor.execute('SELECT checks.status FROM checks INNER JOIN statements ON checks.statement_id = statements.id WHERE statements.line=%d' % line)
        return [row[0] for row in self.cursor.fetchall()]


class TestResult:
    def __init__(self, code, comments=None):
        assert code in ('PASS', 'PASS_IMPROVE', 'FAIL')
        self.code = code
        self.comments = comments or []

    def add_comment(self, comment):
        self.comments.append(comment)


class Test:
    def __init__(self,
                 filename,
                 description,
                 analyses,
                 result,
                 expected=None,
                 domain=None,
                 opt_level=None,
                 entry_points=None,
                 procedural=None,
                 options=None,
                 line_checks=None):
        if not isinstance(analyses, list):
            analyses = [analyses]

        assert all(a in ANALYSES for a in analyses)
        assert result in ('safe', 'unsafe', 'error')
        assert expected in (None, 'safe', 'unsafe', 'error')

        self.filename = filename
        self.description = description
        self.result = result
        self.expected = expected or result
        self.analyses = analyses
        self.domain = domain or 'interval'
        self.opt_level = opt_level or 'basic'
        self.entry_points = entry_points or ('main',)
        self.procedural = procedural or 'inter'
        self.options = options or []
        self.line_checks = line_checks or []

    def run(self, root, output_db):
        fullpath = os.path.join(root, self.filename)
        assert os.path.exists(fullpath)

        # create working directory
        wd = tempfile.mkdtemp(prefix='ikos-%s' % self.filename)
        atexit.register(shutil.rmtree, path=wd)

        # run clang
        bc_path = os.path.join(wd, '%s.bc' % self.filename)
        cmd = [find_clang()]
        cmd += clang_emit_llvm_flags()
        cmd += clang_ikos_flags()
        cmd += [fullpath, '-o', bc_path]
        if self.filename.endswith('.cpp'):
            cmd.append('-std=c++17')
        subprocess.check_call(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)

        # run ikos preprocessor
        pp_path = os.path.join(wd, '%s.pp.bc' % self.filename)
        cmd = [find_ikos_pp(),
               '-opt=%s' % self.opt_level,
               '-entry-points=%s' % ','.join(self.entry_points),
               bc_path,
               '-o', pp_path]
        subprocess.check_call(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)

        # run ikos analyzer
        cmd = [find_ikos_analyzer(),
               '-a=%s' % ','.join(self.analyses),
               '-d=%s' % self.domain,
               '-entry-points=%s' % ','.join(self.entry_points),
               '-proc=%s' % self.procedural]
        cmd.extend(self.options)
        if self.opt_level == 'aggressive':
            cmd.append('-allow-dbg-mismatch')
        if 'gauge' in self.domain:
            cmd.append('-add-loop-counters')
        cmd += [pp_path, '-o', output_db]
        subprocess.check_call(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)

        with Database(output_db) as db:
            # Get the global result
            errors = db.get_num_checks(Result.ERROR)
            warnings = db.get_num_checks(Result.WARNING)

            if errors == 0 and warnings == 0:
                result = 'safe'
            elif errors != 0:
                result = 'error'
            else:
                result = 'unsafe'

            # Set the test result (passed or failed)
            ret = TestResult('PASS')

            if result not in (self.result, self.expected):
                ret.code = 'FAIL'
                ret.add_comment('Got %d errors and %d warnings, was expecting "%s".'
                                % (errors, warnings, self.expected))
            elif result == self.result and result != self.expected:
                ret.code = 'PASS_IMPROVE'
                ret.add_comment('improvement: IKOS returned the right result '
                                '(%s) and not the expected one (%s).'
                                % (self.result, self.expected))

            # Line by line check
            for line_check in self.line_checks:
                if len(line_check) == 3:
                    line_num, line_result, line_expected = line_check
                elif len(line_check) == 2:
                    line_num, line_result = line_check
                    line_expected = line_result
                else:
                    assert False, 'Unexpected line_checks'

                assert line_result in ('ok', 'warning', 'error', 'unreachable')
                assert line_expected in ('ok', 'warning', 'error', 'unreachable')

                result = db.get_line_status(line_num)

                if Result.ERROR in result:
                    result = 'error'
                elif Result.WARNING in result:
                    result = 'warning'
                elif Result.OK in result and all(s in (Result.OK, Result.UNREACHABLE) for s in result):
                    result = 'ok'
                elif result and all(s == Result.UNREACHABLE for s in result):
                    result = 'unreachable'
                else:
                    result = 'unknown'

                if result not in (line_result, line_expected):
                    ret.code = 'FAIL'
                    ret.add_comment('Got status "%s" for line %d, was expecting "%s".'
                                    % (result, line_num, line_expected))
                elif result == line_result and line_result != line_expected:
                    if ret.code == 'PASS':
                        ret.code = 'PASS_IMPROVE'
                    ret.add_comment('improvement: IKOS returned the right result '
                                    '(%s) for line %d and not the expected one (%s).'
                                    % (line_result, line_num, line_expected))

            if ret.code == 'FAIL':
                ret.comments.insert(0, 'Running %r' % cmd)

            return ret


class TestManager:
    def __init__(self, root):
        self.root = root
        self.tests = []
        self.results = {'PASS': 0, 'PASS_IMPROVE': 0, 'FAIL': 0}

    def add(self, test):
        self.tests.append(test)

    def run(self):
        printf(bold('Running tests...\n'))

        # output database
        _, output_db = tempfile.mkstemp(suffix='.db', prefix='ikos-output-')
        atexit.register(lambda: os.unlink(output_db))

        for t in self.tests:
            printf('  %s ... ', t.description)

            if INTERACTIVE:
                printf('\n')
                self.print_progress()
                # Move the cursor right after the '...'
                printf('\r\033[A\033[%dC' % len('  %s ... ' % t.description))

            result = t.run(self.root, output_db)
            self.results[result.code] += 1

            if result.code == 'FAIL':
                printf(red(('Failed\n')))
            elif result.code == 'PASS':
                printf(green('Passed\n'))
            elif result.code == 'PASS_IMPROVE':
                printf(yellow('Passed with improvements!\n'))
            else:
                assert False, 'unknown result'

            if INTERACTIVE:
                # Clear everything down the cursor
                printf('\033[J')

            if VERBOSE:
                for line in result.comments:
                    printf('    %s\n' % line)

        self.print_result()
        exit(self.results['FAIL'])

    def get_num_tests(self):
        return len(self.tests)

    def get_num_done(self):
        return self.results['PASS'] + self.results['PASS_IMPROVE'] + self.results['FAIL']

    def print_result(self):
        printf(bold('Results:\n'))
        if self.results['FAIL'] == 0:
            if self.results['PASS_IMPROVE'] > 0:
                printf(green('  %d tests passed successfully (with %d improvement(s)).\n'
                             % (self.results['PASS'] + self.results['PASS_IMPROVE'],
                                self.results['PASS_IMPROVE'])))
            else:
                printf(green('  %d tests passed successfully.\n' % self.results['PASS']))
        else:
            printf(red('  %d/%d tests failed.\n' % (self.results['FAIL'], self.get_num_tests())))

    def print_progress(self):
        percent = 100.0 * self.get_num_done() / self.get_num_tests()
        full_width = 50
        width = int(full_width * percent / 100.0)

        progressbar = '[' + '#' * width + ' ' * (full_width - width) + '] %d%%' % int(percent)
        printf(progressbar)


def parse_args(description=None):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--verbose', '-v', dest='verbose',
                        help='Verbose mode',
                        action='store_true', default=False)
    parser.add_argument('--no-colors', dest='no_colors',
                        help='Disable colors',
                        action='store_true', default=False)
    parser.add_argument('--no-interactive', dest='no_interactive',
                        help='Disable interactive mode',
                        action='store_true', default=False)
    parser.add_argument('--clang', dest='clang',
                        help='Path to the clang binary',
                        default='clang')
    parser.add_argument('--ikos-pp', dest='ikos_pp',
                        help='Path to the ikos-pp binary',
                        default='ikos-pp')
    parser.add_argument('--ikos-analyzer', dest='ikos_analyzer',
                        help='Path to the ikos-analyzer binary',
                        default='ikos-analyzer')

    args = parser.parse_args()

    global VERBOSE, USE_COLORS, INTERACTIVE, CLANG, IKOS_PP, IKOS_ANALYZER
    VERBOSE = args.verbose
    USE_COLORS = False if args.no_colors else os.isatty(sys.stdout.fileno())
    INTERACTIVE = False if args.no_interactive else os.isatty(sys.stdout.fileno())
    CLANG = args.clang
    IKOS_PP = args.ikos_pp
    IKOS_ANALYZER = args.ikos_analyzer



###############################################################################
#
# Extract information from a result database
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2018-2023 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import collections
import json
import sqlite3

from ikos.enums import FilesTable, FunctionsTable, StatementsTable, \
    CallContextsTable, OperandsTable, MemoryLocationsTable, ChecksTable


class CachedProperty(object):
    ''' A property attribute that only calls its getter the first access. '''

    def __init__(self, factory):
        self._attr_name = factory.__name__
        self._factory = factory

    def __get__(self, instance, owner):
        value = self._factory(instance)
        setattr(instance, self._attr_name, value)
        return value


class OutputDatabase(object):
    ''' Represents an output database '''

    def __init__(self, path):
        self.path = path
        self.con = sqlite3.connect(path)

        # Use 'str' as text factory since it's the type of string literals
        # This is unicode
        self.con.text_factory = str

    def close(self):
        self.con.close()

    def load_settings(self):
        ''' Load the analysis settings from the database '''
        c = self.con.cursor()
        c.execute('SELECT name, value FROM settings')

        settings = {}
        for key, value in c:
            try:
                settings[key] = json.loads(value)
            except ValueError:
                settings[key] = value

        return settings

    def insert_settings(self, rows):
        ''' Insert the analysis settings into the database '''
        c = self.con.cursor()
        c.executemany('INSERT INTO settings VALUES (?, ?)', rows)
        self.con.commit()

    def load_timing_results(self, full=True, sort=True):
        '''
        Load the timing results from the database,
        as a list of tuples (pass, elapsed)
        '''
        c = self.con.cursor()
        where = "WHERE pass NOT LIKE '%ikos-analyzer.%'" if not full else ''
        order_by = 'ORDER BY pass' if sort else ''
        c.execute('SELECT pass, time FROM times %s %s' % (where, order_by))
        return c.fetchall()

    def insert_timing_results(self, rows):
        ''' Insert the timing results into the database '''
        c = self.con.cursor()
        c.executemany('INSERT INTO times VALUES (?, ?)', rows)
        self.con.commit()

    @CachedProperty
    def files(self):
        return self._fetch_table('files', File)

    @CachedProperty
    def functions(self):
        return self._fetch_table('functions', Function)

    @CachedProperty
    def statements(self):
        return self._fetch_table('statements', Statement)

    @CachedProperty
    def operands(self):
        return self._fetch_table('operands', Operand)

    @CachedProperty
    def call_contexts(self):
        return self._fetch_table('call_contexts', CallContext)

    @CachedProperty
    def memory_locations(self):
        return self._fetch_table('memory_locations', MemoryLocation)

    def _fetch_table(self, table, klass):
        c = self.con.cursor()
        c.execute('SELECT * FROM %s ORDER BY id' % table)
        return [klass(row, self) for row in c]


class File(object):
    ''' Represents a source file '''

    __slots__ = ('id', 'path')

    def __init__(self, row, db):
        self.id = row[FilesTable.ID]
        self.path = row[FilesTable.PATH]


class Function(object):
    ''' Represents a function '''

    __slots__ = (
        'id',
        'name',
        'demangled',
        'definition',
        'file_id',
        'line',
        'db'
    )

    def __init__(self, row, db):
        self.id = row[FunctionsTable.ID]
        self.name = row[FunctionsTable.NAME]
        self.demangled = row[FunctionsTable.DEMANGLED]  # or None
        self.definition = (row[FunctionsTable.DEFINITION] == 1)
        self.file_id = row[FunctionsTable.FILE_ID]  # or None
        self.line = row[FunctionsTable.LINE]  # or None
        self.db = db

    def pretty_name(self):
        ''' Return a pretty name '''
        if self.demangled:
            return self.demangled
        else:
            return self.name

    def file(self):
        ''' Return the source file, or None '''
        if self.file_id is None:
            return None

        return self.db.files[self.file_id]

    def file_path(self):
        ''' Return the source file path, or None '''
        if self.file_id is None:
            return None

        return self.db.files[self.file_id].path


class Statement(object):
    ''' Represents a statement '''

    __slots__ = (
        'id',
        'kind',
        'function_id',
        'file_id',
        'line',
        'column',
        'db'
    )

    def __init__(self, row, db):
        self.id = row[StatementsTable.ID]
        self.kind = row[StatementsTable.KIND]
        self.function_id = row[StatementsTable.FUNCTION_ID]
        self.file_id = row[StatementsTable.FILE_ID]  # or None
        self.line = row[StatementsTable.LINE]  # or None
        self.column = row[StatementsTable.COLUMN]  # or None
        self.db = db

    def function(self):
        ''' Return the function '''
        return self.db.functions[self.function_id]

    def file(self):
        ''' Return the source file, or None '''
        if self.file_id is None:
            return None

        return self.db.files[self.file_id]

    def file_path(self):
        ''' Return the source file path, or None '''
        if self.file_id is None:
            return None

        return self.db.files[self.file_id].path

    def file_id_or(self, default):
        ''' Return the file id, or default '''
        if self.file_id is None:
            return default

        return self.file_id

    def line_or(self, default):
        ''' Return the line number, or default '''
        if self.line is None:
            return default

        return self.line

    def column_or(self, default):
        ''' Return the column number, or None '''
        if self.column is None:
            return default

        return self.column


class Operand(object):
    ''' Represents an operand '''

    __slots__ = ('id', 'kind', 'repr')

    def __init__(self, row, db):
        self.id = row[OperandsTable.ID]
        self.kind = row[OperandsTable.KIND]
        self.repr = row[OperandsTable.REPR]


# Represents a pair (operand number, operand)
NumOperandPair = collections.namedtuple('NumOperandPair', ('num', 'operand'))


class CallContext(object):
    ''' Represents a calling context '''

    __slots__ = ('id', 'call_id', 'function_id', 'parent_id', 'db')

    def __init__(self, row, db):
        self.id = row[CallContextsTable.ID]
        self.call_id = row[CallContextsTable.CALL_ID]  # or None
        self.function_id = row[CallContextsTable.FUNCTION_ID]  # or None
        self.parent_id = row[CallContextsTable.PARENT_ID]  # or None
        self.db = db

    def empty(self):
        return self.call_id is None

    def call(self):
        ''' Return the call statement '''
        assert self.call_id is not None
        return self.db.statements[self.call_id]

    def function(self):
        ''' Return the function '''
        assert self.function_id is not None
        return self.db.functions[self.function_id]

    def parent(self):
        ''' Return the parent calling context '''
        assert self.parent_id is not None
        return self.db.call_contexts[self.parent_id]

    def str(self):
        ''' Format a calling context '''
        call_context = self
        ctx = []

        while not call_context.empty():
            function = call_context.function()
            call = call_context.call()

            ctx.append('%s@%s:%s:%d' % (function.pretty_name(),
                                        call.line_or('?'),
                                        call.column_or('?'),
                                        call_context.call_id))

            call_context = call_context.parent()

        ctx.append('.')
        ctx.reverse()
        return '/'.join(ctx)

    def __str__(self):
        ''' Format a calling context '''
        return self.str()


class MemoryLocation(object):
    ''' Represents a memory location '''

    __slots__ = ('id', 'kind', 'info', 'db')

    def __init__(self, row, db):
        self.id = row[MemoryLocationsTable.ID]
        self.kind = row[MemoryLocationsTable.KIND]
        self.info = row[MemoryLocationsTable.INFO]  # or None
        self.db = db

    def load_info(self):
        ''' Return the info, or None '''
        if not self.info:
            return None

        return json.loads(self.info)


class Check(object):
    ''' Represents a check '''

    __slots__ = (
        'id',
        'kind',
        'checker',
        'status',
        'statement_id',
        'call_context_id',
        'operands',
        'info',
        'db'
    )

    def __init__(self, row, db):
        self.id = row[ChecksTable.ID]
        self.kind = row[ChecksTable.KIND]
        self.checker = row[ChecksTable.CHECKER]
        self.status = row[ChecksTable.STATUS]
        self.statement_id = row[ChecksTable.STATEMENT_ID]
        self.call_context_id = row[ChecksTable.CALL_CONTEXT_ID]
        self.operands = row[ChecksTable.OPERANDS]  # or None
        self.info = row[ChecksTable.INFO]  # or None
        self.db = db

    def statement(self):
        ''' Return the statement '''
        return self.db.statements[self.statement_id]

    def call_context(self):
        ''' Return the call context '''
        return self.db.call_contexts[self.call_context_id]

    def load_operands(self):
        ''' Return the operands, or None '''
        if not self.operands:
            return None

        operands = json.loads(self.operands)
        return [NumOperandPair(num, self.db.operands[id])
                for num, id in operands]



###############################################################################
#
# IKOS web view
#
# Author: Thomas Bailleux
#
# Contributor: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2018-2023 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import argparse
import collections
import html
import io
import json
import operator
import os
import os.path
import re
import sqlite3
import sys
import threading
import webbrowser

from ikos import args
from ikos import colors
from ikos import log
from ikos import report
from ikos import settings
from ikos.enums import Result, CheckKind
from ikos.highlight import CppLexer, HtmlFormatter, highlight
from ikos.http import HTTPServerIPv6, BaseHTTPRequestHandler
from ikos.log import printf
from ikos.output_db import OutputDatabase


# Path to web resources
SHARE_DIR = os.path.join(settings.PREFIX, 'share', 'ikos', 'view')


class TemplateEngine:
    ''' Simple template engine for formatting HTML page '''

    _singleton = None

    @classmethod
    def get(cls):
        ''' Get the singleton instance of the template engine '''
        if cls._singleton is None:
            cls._singleton = TemplateEngine()

        return cls._singleton

    def __init__(self):
        self._cache = {}

    def _read(self, fullpath):
        ''' Read a template from a file or cache '''
        if fullpath in self._cache and settings.BUILD_MODE == 'Release':
            log.debug("Using '%s' from cache" % fullpath)
            return self._cache[fullpath]
        else:
            with io.open(fullpath, 'r',
                         encoding='utf-8',
                         errors='ignore') as f:
                data = f.read()
            self._cache[fullpath] = data
            return data

    def process(self, path, values={}):
        '''
        Format an HTML page with a given template and a dictionary of values
        '''
        # read the template file
        fullpath = os.path.join(SHARE_DIR, 'template', path)
        data = self._read(fullpath)

        # decode bytes as utf-8
        for key in values:
            if isinstance(values[key], bytes):
                values[key] = values[key].decode('utf-8')

        # substitute {xxx} placeholders
        try:
            data = data.format(**values)
        except KeyError:
            pass

        return data


class RequestHandler(BaseHTTPRequestHandler):
    ''' Handler for an HTTP request '''

    _static_cache = {}

    def do_GET(self):
        ''' Handle a GET request '''
        urls = [
            (r'^/$',
             self._serve_homepage),
            (r'^/static/(?P<path>[a-zA-Z_-]+/[a-zA-Z0-9_-]+\.[a-zA-Z_-]+)$',
             self._serve_static),
            (r'^/settings$',
             self._serve_settings),
            (r'^/report/(?P<id>[0-9]+)(\?k=(?P<kinds_filter>[A-Z0-9]+))?$',
             self._serve_report)
        ]

        for pattern, f in urls:
            m = re.match(pattern, self.path)
            if m:
                f(**m.groupdict())
                return

        self._serve_not_found()

    # Static files

    def _serve_static(self, path):
        ''' Serve a static file '''
        fullpath = os.path.join(SHARE_DIR, 'static', path)

        if os.path.isfile(fullpath):
            self._send_static_headers(path)
            self._write_file(fullpath)
        else:
            self._serve_not_found()

    # Homepage

    def _serve_homepage(self):
        ''' Serve the homepage '''
        self._write_template('homepage.html', {
            'check_kinds': json.dumps(self._check_kinds()),
            'check_kinds_filter': json.dumps(self._check_kinds_filter()),
            'files': json.dumps(self._files()),
        })

    def _check_kinds(self):
        ''' Generate the Javascript variable check_kinds '''
        view_report = View.get().report
        return [{'id': kind, 'name': CheckKind.long_name(kind)}
                for kind in view_report.kinds]

    def _check_kinds_filter(self, param=None):
        ''' Generate the Javascript variable check_kinds_filter '''
        view_report = View.get().report

        filter = {kind: True for kind in view_report.kinds}

        if not param:
            return filter

        param = [int(param[i:i + 2], 16) for i in range(0, len(param), 2)]

        for kind in view_report.kinds:
            try:
                filter[kind] = (param[kind // 8] & (1 << (kind % 8))) != 0
            except IndexError:
                pass

        return filter

    def _files(self):
        ''' Generate the Javascript variable files '''
        view_report = View.get().report
        files = []

        for file in view_report.files:
            files.append({
                'id': file.id,
                'path': report.format_path(file.path),
                'status_kinds': view_report.files_status_kinds[file.id]
            })

        # Sort by path
        files.sort(key=operator.itemgetter('path'))

        return files

    # Settings

    def _serve_settings(self):
        ''' Serve the settings page '''
        db = View.get().db
        settings = db.load_settings()
        s = []

        for name, value in settings.items():
            if isinstance(value, list):
                s.append('<tr><td>%s</td><td><i>%s</i></td></tr>'
                         % (html.escape(name),
                            html.escape(json.dumps(value))))
            else:
                s.append('<tr><td>%s</td><td><i>%s</i></td></tr>'
                         % (html.escape(name),
                            html.escape(str(value))))

        self._write_template('settings.html', {'settings': '\n\t'.join(s)})

    # File report

    def _serve_report(self, id, kinds_filter):
        ''' Serve a specific report for a file '''
        id = int(id)
        view_report = View.get().report

        try:
            file = view_report.files[id]
        except IndexError:
            self._serve_not_found()
            return

        try:
            with io.open(file.path, 'r',
                         encoding='utf-8',
                         errors='ignore') as f:
                code = f.read()
        except (OSError, IOError):
            self._serve_error("No such file: %s" % file.path)
            return

        fmt = Formatter(file)
        lexer = CppLexer(stripnl=False)
        code = highlight(code, lexer, fmt)
        check_kinds_filter = self._check_kinds_filter(param=kinds_filter)
        self._write_template('report.html', {
            'filepath': html.escape(report.format_path(file.path)),
            'check_kinds': json.dumps(self._check_kinds()),
            'check_kinds_filter': json.dumps(check_kinds_filter),
            'code': code,
            'functions': json.dumps(fmt.functions),
            'call_contexts': json.dumps(fmt.call_contexts),
            'checks': json.dumps(fmt.checks),
            'pygments_css': fmt.get_style_defs('.highlight')
        })

    # Helpers

    def _send_static_headers(self, path):
        ''' Send header for a static file '''
        self.send_response(200)
        mimetype = 'text/plain'
        if path.endswith('.css'):
            mimetype = 'text/css'
        elif path.endswith('.js'):
            mimetype = 'application/javascript'
        elif path.endswith('.jpg'):
            mimetype = 'image/jpg'
        elif path.endswith('.png'):
            mimetype = 'image/png'
        self.send_header('Content-Type', '%s; charset=UTF-8' % mimetype)
        self.end_headers()

    def _write_file(self, fullpath):
        ''' Write a static file to the response stream '''
        if fullpath in RequestHandler._static_cache \
                and settings.BUILD_MODE == 'Release':
            self.wfile.write(RequestHandler._static_cache[fullpath])
            log.debug("Using '%s' from cache" % fullpath)
        else:
            with open(fullpath, 'rb') as f:
                data = f.read()
            self.wfile.write(data)
            RequestHandler._static_cache[fullpath] = data

    def _write_template(self, path, values={}, status=200):
        ''' Write a template to the response stream '''
        engine = TemplateEngine.get()
        self.send_response(status)
        self.send_header('Content-Type', 'text/html; charset=UTF-8')
        self.end_headers()
        self.wfile.write(engine.process(path, values).encode('utf8'))

    def _serve_not_found(self):
        ''' Server a 404 Not Found page '''
        self._write_template('not_found.html',
                             {'path': html.escape(self.path)},
                             status=404)

    def _serve_error(self, message):
        ''' Server a 500 Internal Error page '''
        self._write_template('error.html',
                             {'message': html.escape(message)},
                             status=500)


class View:
    ''' Class for IKOS view (ikos-view) '''

    _singleton = None

    @classmethod
    def get(cls):
        ''' Get the singleton instance of view '''
        assert cls._singleton is not None
        return cls._singleton

    def __init__(self, db, port=8080):
        View._singleton = self
        self.db = db
        self.port = port
        self.report = ViewReport(self.db)

        try:
            self.httpd = HTTPServerIPv6(('', self.port), RequestHandler)
        except (OSError, IOError) as e:
            log.error("Could not start the HTTP server: %s" % e)
            sys.exit(1)

    def serve(self):
        log.info("Pre-processing...")
        self.report.pre_process()
        log.info("Starting ikos-view at http://localhost:%d" % self.port)
        log.warning("Use Ctrl-C to exit")
        self.httpd.serve_forever()


StatusKinds = collections.namedtuple('StatusKinds',
                                     ['ok', 'warning', 'error', 'unreachable'])


class ViewReport:
    ''' IKOS view report '''

    def __init__(self, db):
        self.db = db
        self._report = None
        self.kinds = None
        self.files = None

        # Map[file.id, Map[check.status, Map[check.kind, count]]]
        self.files_status_kinds = None

        # Map[file.id, Map[check.line, List[Check]]]
        self.files_lines_reports = None

    def pre_process(self):
        ''' Pre processing some values '''
        # List of CheckKind
        c = self.db.con.cursor()
        rows = c.execute("SELECT DISTINCT kind FROM checks ORDER BY kind")
        self.kinds = [row[0] for row in rows]

        # Generate report
        self._report = report.generate_report(self.db)

        self.files = self.db.files
        self.files_status_kinds = {}
        self.files_lines_reports = {}
        for file in self.files:
            self.files_status_kinds[file.id] = StatusKinds(ok={},
                                                           warning={},
                                                           error={},
                                                           unreachable={})
            self.files_lines_reports[file.id] = {}

        for statement_report in self._report.statement_reports:
            stmt = statement_report.statement()
            file = stmt.file()

            if file is None:
                continue  # Ignore checks without a file

            kind = statement_report.kind
            status = statement_report.status

            # Update self.files_status_kinds
            kinds = self.files_status_kinds[file.id][status]
            if kind not in kinds:
                kinds[kind] = 0
            kinds[kind] += 1

            # Update self.files_lines_reports
            lines_reports = self.files_lines_reports[file.id]
            if stmt.line not in lines_reports:
                lines_reports[stmt.line] = []
            lines_reports[stmt.line].append(statement_report)


class Formatter(HtmlFormatter):
    ''' Source code HTML formatter '''

    def __init__(self, file):
        super(Formatter, self).__init__()
        self.file = file
        self.functions = {}
        self.call_contexts = {}
        self.checks = {}

    def wrap(self, source):
        return self._wrap_code(source)

    def _wrap_code(self, source):
        view_report = View.get().report
        lines_reports = view_report.files_lines_reports[self.file.id]

        yield 0, '<div class="highlight">\n'
        line_num = 1

        for i, line in source:
            statement_reports = lines_reports.get(line_num)

            t = '''<div id="L%d" class="line_wrap status_%d">
                     <span class="line_number">
                      <a href="#L%d">%d</a>
                     </span>
                     <span class="line_content">
                       <span class="line_code_wrap">
                         <span class="line_code">
                           <pre>%s</pre>
                         </span>
                         <span class="checks_toggle hidden">
                           <a href="javascript:" class="toggle">&#128065;</a>
                         </span>
                       </span>
                       <span class="checks hidden"></span>
                     </span>
                   </div>\n''' % (line_num,
                                  self._line_status(statement_reports),
                                  line_num,
                                  line_num,
                                  line.replace('\n', ''))

            if statement_reports:
                self.checks[line_num] = self._build_checks(statement_reports)

            yield i, t
            line_num += 1

        yield 0, '</div>'

    def _line_status(self, statement_reports):
        ''' Return the status of a source line '''
        if not statement_reports:
            return -1

        status = [0, 0, 0, 0]
        for statement_report in statement_reports:
            status[statement_report.status] += 1

        if status[Result.ERROR] > 0:
            return Result.ERROR
        elif status[Result.WARNING] > 0:
            return Result.WARNING
        elif status[Result.UNREACHABLE] > 0:
            return Result.UNREACHABLE
        else:
            return Result.OK

    def _build_checks(self, statement_reports):
        ''' Return the list of check for a source line '''
        checks = []

        for statement_report in statement_reports:
            statement = statement_report.statement()
            checks.append({
                'kind': statement_report.kind,
                'status': statement_report.status,
                'column': statement.column_or('?'),
                'message': report.generate_message(statement_report, 4),
                'function_id': statement.function_id,
                'call_context_ids': statement_report.call_context_ids,
            })
            self._build_function(statement.function())
            self._build_call_contexts(statement_report.call_contexts())

        # Sort by status (error > warning > ok)
        checks.sort(key=operator.itemgetter('status'), reverse=True)

        return checks

    def _build_function(self, function):
        if function.id in self.functions:
            return

        self.functions[function.id] = function.pretty_name()

    def _build_call_contexts(self, call_contexts):
        for call_context in call_contexts:
            self._build_call_context(call_context)

    def _build_call_context(self, call_context):
        call_context_id = call_context.id

        if call_context_id in self.call_contexts:
            return

        lines = []
        while not call_context.empty():
            call_statement = call_context.call()
            function = call_context.function()
            line = "%s:%s:%s: function '%s'" % (
                report.format_path(call_statement.file_path()) or '?',
                call_statement.line_or('?'),
                call_statement.column_or('?'),
                function.pretty_name()
            )
            lines.append(line)
            call_context = call_context.parent()

        self.call_contexts[call_context_id] = '\n'.join(lines)


##########################
# command line interface #
##########################


def parse_arguments(argv):
    usage = '%(prog)s [options] file.db'
    description = 'Launch ikos view on an output database'
    formatter_class = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(usage=usage,
                                     description=description,
                                     formatter_class=formatter_class)

    # Positional arguments
    parser.add_argument('file',
                        metavar='file.db',
                        help='Result database')

    # Optional arguments
    parser.add_argument('--version',
                        action=args.VersionAction,
                        nargs=0,
                        help='Show ikos version')
    parser.add_argument('--color',
                        dest='color',
                        metavar='',
                        help=args.help('Enable terminal colors:',
                                       args.color_choices,
                                       args.default_color),
                        choices=args.choices(args.color_choices),
                        default=args.default_color)
    parser.add_argument('--log',
                        dest='log_level',
                        metavar='',
                        help=args.help('Log level:',
                                       args.log_levels,
                                       args.default_log_level),
                        choices=args.choices(args.log_levels),
                        default='info')
    parser.add_argument('--port',
                        dest='port',
                        metavar='',
                        help='Listening port',
                        default=8080,
                        type=int)

    return parser.parse_args(argv)


def open_browser(url):
    browser = webbrowser.get()

    if browser.name in ('links', 'elinks', 'lynx', 'w3m'):
        return

    log.debug('Launching browser')
    browser.open_new_tab(url)


######################
# main for ikos-view #
######################

def main(argv):
    progname = os.path.basename(argv[0])

    # parse arguments
    opt = parse_arguments(argv[1:])

    # setup colors and logging
    colors.setup(opt.color, file=log.out)
    log.setup(opt.log_level)

    if not os.path.isfile(opt.file):
        printf("%s: error: no such file: \'%s\'\n",
               progname, opt.file, file=sys.stderr)
        sys.exit(1)

    try:
        # open result database
        db = OutputDatabase(opt.file)

        v = View(db, port=opt.port)
        browser_timer = threading.Timer(0.1,
                                        open_browser,
                                        ['http://localhost:%d/' % opt.port])
        browser_timer.start()
        v.serve()

        # close database
        db.close()
    except sqlite3.DatabaseError as e:
        printf('%s: error: %s\n', progname, e, file=sys.stderr)
        sys.exit(1)



###############################################################################
#
# Handle ikos toolchain: running clang, ikos-pp and ikos-analyzer.
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2011-2019 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import argparse
import atexit
import datetime
import json
import os
import os.path
import shlex
import shutil
import signal
import subprocess
import sys
import tempfile
import threading

from ikos import args
from ikos import colors
from ikos import log
from ikos import report
from ikos import settings
from ikos import stats
from ikos.log import printf
from ikos.output_db import OutputDatabase


def parse_arguments(argv):
    usage = '%(prog)s [options] file[.c|.cpp|.bc|.ll]'
    description = 'ikos static analyzer'
    formatter_class = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(usage=usage,
                                     description=description,
                                     formatter_class=formatter_class)

    # Positional arguments
    parser.add_argument('file',
                        metavar='file[.c|.cpp|.bc|.ll]',
                        help='File to analyze')

    # Optional arguments
    parser.add_argument('-o', '--output-db',
                        dest='output_db',
                        metavar='<file>',
                        help='Output database file (default: output.db)',
                        default='output.db')
    parser.add_argument('-v',
                        dest='verbosity',
                        help='Increase verbosity',
                        action='count',
                        default=1)
    parser.add_argument('-q',
                        dest='verbosity',
                        help='Be quiet',
                        action='store_const',
                        const=0)
    parser.add_argument('--version',
                        action=args.VersionAction,
                        nargs=0,
                        help='Show ikos version')

    # Analysis options
    analysis = parser.add_argument_group('Analysis Options')
    analysis.add_argument('-a', '--analyses',
                          dest='analyses',
                          metavar='',
                          help=args.help('Available analyses:',
                                         args.analyses,
                                         args.default_analyses),
                          action='append')
    analysis.add_argument('-d', '--domain',
                          dest='domain',
                          metavar='',
                          help=args.help('Available abstract domains:',
                                         args.domains,
                                         args.default_domain),
                          choices=args.choices(args.domains),
                          default=args.default_domain)
    analysis.add_argument('-e', '--entry-points',
                          dest='entry_points',
                          metavar='<function>',
                          help='List of program entry points (default: main)',
                          action='append')
    analysis.add_argument('--globals-init',
                          dest='globals_init',
                          metavar='',
                          help=args.help(
                              'Policy of initialization for global variables:',
                              args.globals_init_policies,
                              args.default_globals_init_policy),
                          choices=args.choices(args.globals_init_policies),
                          default=args.default_globals_init_policy)
    analysis.add_argument('--no-init-globals',
                          dest='no_init_globals',
                          metavar='<function>',
                          help='Do not initialize global variables for the '
                               'given entry points',
                          action='append')
    analysis.add_argument('--no-liveness',
                          dest='no_liveness',
                          help='Disable the liveness analysis',
                          action='store_true',
                          default=False)
    analysis.add_argument('--no-pointer',
                          dest='no_pointer',
                          help='Disable the pointer analysis',
                          action='store_true',
                          default=False)
    analysis.add_argument('--no-widening-hints',
                          dest='no_widening_hints',
                          help='Disable the widening hint analysis',
                          action='store_true',
                          default=False)
    analysis.add_argument('--no-fixpoint-cache',
                          dest='no_fixpoint_cache',
                          help='Disable the cache of fixpoints',
                          action='store_true',
                          default=False)
    analysis.add_argument('--no-checks',
                          dest='no_checks',
                          help='Disable all the checks',
                          action='store_true',
                          default=False)
    analysis.add_argument('--proc',
                          dest='procedural',
                          metavar='',
                          help=args.help('Procedural:',
                                         args.proceduralities,
                                         args.default_procedurality),
                          choices=args.choices(args.proceduralities),
                          default=args.default_procedurality)
    analysis.add_argument('-j', '--jobs',
                          dest='jobs',
                          metavar='',
                          nargs='?',
                          help='Number of threads',
                          type=int,
                          const=0,
                          default=1)
    analysis.add_argument('--widening-strategy',
                          dest='widening_strategy',
                          metavar='',
                          help=args.help('Strategy for increasing iterations:',
                                         args.widening_strategies,
                                         args.default_widening_strategy),
                          choices=args.choices(args.widening_strategies),
                          default=args.default_widening_strategy)
    analysis.add_argument('--narrowing-strategy',
                          dest='narrowing_strategy',
                          metavar='',
                          help=args.help('Strategy for decreasing iterations:',
                                         args.narrowing_strategies,
                                         args.default_narrowing_strategy),
                          choices=args.choices(args.narrowing_strategies),
                          default=args.default_narrowing_strategy)
    analysis.add_argument('--widening-delay',
                          dest='widening_delay',
                          metavar='',
                          help='Number of loop iterations before using the'
                               ' widening strategy (default: %d)'
                               % args.default_widening_delay,
                          default=args.default_widening_delay,
                          type=args.Integer(min=0))
    analysis.add_argument('--widening-delay-functions',
                          dest='widening_delay_functions',
                          metavar='<function:int>',
                          help='Widening delay for specific functions',
                          action='append')
    analysis.add_argument('--widening-period',
                          dest='widening_period',
                          metavar='',
                          help='Number of loop iterations between each'
                               ' widening (default: %d)'
                               % args.default_widening_period,
                          default=args.default_widening_period,
                          type=args.Integer(min=1))
    analysis.add_argument('--narrowing-iterations',
                          dest='narrowing_iterations',
                          metavar='',
                          help='Perform a fixed number of narrowing'
                               ' iterations',
                          type=args.Integer(min=0))
    analysis.add_argument('--partitioning',
                          dest='partitioning',
                          metavar='',
                          help=args.help('Partitioning strategy:',
                                         args.partitioning_strategies,
                                         args.default_partitioning_strategy),
                          choices=args.choices(args.partitioning_strategies),
                          default=args.default_partitioning_strategy)
    analysis.add_argument('--hardware-addresses',
                          dest='hardware_addresses',
                          metavar='',
                          action='append',
                          help='Specify ranges (x-y) of hardware addresses'
                               ', separated by a comma')
    analysis.add_argument('--hardware-addresses-file',
                          dest='hardware_addresses_file',
                          metavar='',
                          help='Specify ranges (x-y) of hardware addresses'
                               ' from a file (one range per line)')
    analysis.add_argument('--argc',
                          dest='argc',
                          metavar='',
                          help='Specify a value for argc',
                          type=args.Integer(min=0))

    # Compile options
    compiler = parser.add_argument_group('Compile Options')
    compiler.add_argument('-I',
                          dest='compiler_include_flags',
                          metavar='',
                          help='Add the specified directory to the search '
                               'path for include files',
                          action='append')
    compiler.add_argument('-D',
                          dest='compiler_define_flags',
                          metavar='',
                          help='Add an implicit #define into the source file',
                          action='append')
    compiler.add_argument('-W',
                          dest='compiler_warning_flags',
                          metavar='',
                          help='Use the specified warning options',
                          action='append')
    compiler.add_argument('-w',
                          dest='compiler_disable_warnings',
                          help='Suppress all compiler warnings',
                          action='store_true',
                          default=False)
    compiler.add_argument('-m',
                          dest='compiler_machine_flags',
                          metavar='',
                          help='Use the specified machine options',
                          action='append')

    # Preprocessing options
    preprocess = parser.add_argument_group('Preprocessing Options')
    preprocess.add_argument('--opt',
                            dest='opt_level',
                            metavar='',
                            help=args.help('Optimization level:',
                                           args.opt_levels,
                                           args.default_opt_level),
                            choices=args.choices(args.opt_levels),
                            default=args.default_opt_level)
    preprocess.add_argument('--inline-all',
                            dest='inline_all',
                            help='Front-end inline all functions',
                            action='store_true',
                            default=False)
    preprocess.add_argument('--no-bc-verify',
                            dest='no_bc_verify',
                            help='Do not run the LLVM bitcode verifier',
                            action='store_true',
                            default=False)

    # Import options
    imports = parser.add_argument_group('Import Options')
    imports.add_argument('--no-libc',
                         dest='no_libc',
                         help='Do not use libc intrinsics '
                              '(malloc, free, etc.)',
                         action='store_true',
                         default=False)
    imports.add_argument('--no-libcpp',
                         dest='no_libcpp',
                         help='Do not use libcpp intrinsics '
                              '(__cxa_throw, etc.)',
                         action='store_true',
                         default=False)
    imports.add_argument('--no-libikos',
                         dest='no_libikos',
                         help='Do not use ikos intrinsics '
                              '(__ikos_assert, etc.)',
                         action='store_true',
                         default=False)

    # AR passes options
    passes = parser.add_argument_group('AR Passes Options')
    passes.add_argument('--no-type-check',
                        dest='no_type_check',
                        help='Do not run the AR type checker',
                        action='store_true',
                        default=False)
    passes.add_argument('--no-simplify-cfg',
                        dest='no_simplify_cfg',
                        help='Do not run the simplify-cfg pass',
                        action='store_true',
                        default=False)
    passes.add_argument('--no-simplify-upcast-comparison',
                        dest='no_simplify_upcast_comparison',
                        help='Do not run the simplify-upcast-comparison pass',
                        action='store_true',
                        default=False)

    # Debug options
    debug = parser.add_argument_group('Debug Options')
    debug.add_argument('--display-llvm',
                       dest='display_llvm',
                       help='Display the LLVM bitcode as text',
                       action='store_true',
                       default=False)
    debug.add_argument('--display-ar',
                       dest='display_ar',
                       help='Display the Abstract Representation as text',
                       action='store_true',
                       default=False)
    debug.add_argument('--trace-ar-stmts',
                       dest='trace_ar_stmts',
                       help='Trace analysis of ar statements',
                       action='store_true',
                       default=False)
    debug.add_argument('--display-liveness',
                       dest='display_liveness',
                       help='Display liveness analysis results',
                       action='store_true',
                       default=False)
    debug.add_argument('--display-function-pointer',
                       dest='display_function_pointer',
                       help='Display function pointer analysis results',
                       action='store_true',
                       default=False)
    debug.add_argument('--display-pointer',
                       dest='display_pointer',
                       help='Display pointer analysis results',
                       action='store_true',
                       default=False)
    debug.add_argument('--display-fixpoint-parameters',
                       dest='display_fixpoint_parameters',
                       help='Display fixpoint parameters',
                       action='store_true',
                       default=False)
    debug.add_argument('--display-checks',
                       dest='display_checks',
                       metavar='',
                       help=args.help('Display checks:',
                                      args.display_checks_choices,
                                      'no'),
                       choices=args.choices(args.display_checks_choices),
                       default='no')
    debug.add_argument('--display-inv',
                       dest='display_inv',
                       metavar='',
                       help=args.help('Display computed invariants:',
                                      args.display_inv_choices,
                                      'no'),
                       choices=args.choices(args.display_inv_choices),
                       default='no')
    debug.add_argument('--display-raw-checks',
                       dest='display_raw_checks',
                       help='Display analysis raw checks',
                       action='store_true',
                       default=False)
    debug.add_argument('--generate-dot',
                       dest='generate_dot',
                       help='Generate a .dot file for each function',
                       action='store_true',
                       default=False)
    debug.add_argument('--generate-dot-dir',
                       dest='generate_dot_dir',
                       metavar='<directory>',
                       help='Output directory for .dot files',
                       default=None)
    debug.add_argument('--save-temps',
                       dest='save_temps',
                       help='Do not delete temporary files',
                       action='store_true',
                       default=False)
    debug.add_argument('--temp-dir',
                       dest='temp_dir',
                       metavar='<directory>',
                       help='Temporary directory',
                       default=None)

    # Misc.
    misc = parser.add_argument_group('Miscellaneous')
    misc.add_argument('--rm-db',
                      dest='remove_db',
                      help='Remove the output database file after use',
                      action='store_true',
                      default=False)
    misc.add_argument('--color',
                      dest='color',
                      metavar='',
                      help=args.help('Enable terminal colors:',
                                     args.color_choices,
                                     args.default_color),
                      choices=args.choices(args.color_choices),
                      default=args.default_color)
    misc.add_argument('--log',
                      dest='log_level',
                      metavar='',
                      help=args.help('Log level:',
                                     args.log_levels,
                                     args.default_log_level),
                      choices=args.choices(args.log_levels),
                      default=None)
    misc.add_argument('--progress',
                      dest='progress',
                      metavar='',
                      help=args.help('Progress report:',
                                     args.progress_choices,
                                     args.default_progress),
                      choices=args.choices(args.progress_choices),
                      default=args.default_progress)

    # Report options
    report = parser.add_argument_group('Report Options')
    report.add_argument('--display-times',
                        dest='display_times',
                        metavar='',
                        help=args.help('Display timing results',
                                       args.display_times_choices,
                                       'short'),
                        choices=args.choices(args.display_times_choices),
                        default='short')
    report.add_argument('--display-summary',
                        dest='display_summary',
                        metavar='',
                        help=args.help('Display the analysis summary',
                                       args.display_summary_choices,
                                       'full'),
                        choices=args.choices(args.display_summary_choices),
                        default='full')
    report.add_argument('-f', '--format',
                        dest='format',
                        metavar='',
                        help=args.help('Available report formats:',
                                       args.report_formats,
                                       'auto'),
                        choices=args.choices(args.report_formats),
                        default='auto')
    report.add_argument('--report-file',
                        dest='report_file',
                        metavar='<file>',
                        help='Write the report into a file (default: stdout)',
                        default=sys.stdout,
                        type=argparse.FileType('w'))
    report.add_argument('--status-filter',
                        dest='status_filter',
                        metavar='',
                        help=args.help('Available status filters:',
                                       args.status_filters,
                                       args.default_status_filter),
                        action='append')
    report.add_argument('--report-verbosity',
                        dest='report_verbosity',
                        metavar='[1-4]',
                        help='Report verbosity (default: 1)',
                        type=args.Integer(min=1, max=4))

    # Resource options
    resource = parser.add_argument_group('Resources Options')
    resource.add_argument('--cpu',
                          dest='cpu',
                          help='CPU time limit (seconds)',
                          type=args.Integer(min=1))
    resource.add_argument('--mem',
                          dest='mem',
                          help='MEM limit (MB)',
                          type=args.Integer(min=1))

    opt = parser.parse_args(argv)

    # parse --analyses
    opt.analyses = args.parse_argument(parser,
                                       'analyses',
                                       choices=args.analyses,
                                       groups=None,
                                       default=args.default_analyses,
                                       value=opt.analyses)

    # by default, the entry point is main
    if not opt.entry_points:
        opt.entry_points = ('main',)

    # verbosity changes the log level, if --log is not specified
    if opt.log_level is None:
        if opt.verbosity <= 0:
            opt.log_level = 'error'
        elif opt.verbosity == 1:
            opt.log_level = 'info'
        elif opt.verbosity == 2:
            opt.log_level = 'debug'
        else:
            opt.log_level = 'all'

    # quiet mode does not display the timing results and the analysis summary
    if opt.verbosity <= 0:
        opt.display_times = 'no'
        opt.display_summary = 'no'

    # default value for generate-dot-dir
    if opt.generate_dot and not opt.generate_dot_dir:
        if opt.temp_dir and opt.save_temps:
            opt.generate_dot_dir = opt.temp_dir
        else:
            opt.generate_dot_dir = '.'

    # parse --status-filter
    opt.status_filter = args.parse_argument(parser,
                                            'status-filter',
                                            choices=args.status_filters,
                                            groups=None,
                                            default=args.default_status_filter,
                                            value=opt.status_filter)

    # verbosity changes the report verbosity level,
    # if --report-verbosity is not specified
    if opt.report_verbosity is None:
        opt.report_verbosity = max(opt.verbosity, 1)

    return opt


sh_quote = shlex.quote


def command_string(cmd):
    return ' '.join(map(sh_quote, cmd))


def path_ext(path):
    ''' Return the filename extension.

    >>> path_ext('test.c')
    '.c'
    >>> path_ext('/tmp/file.my.ext')
    '.ext'
    '''
    return os.path.splitext(path)[1]


c_extensions = ('.c', '.h', '.i')
cpp_extensions = ('.cpp', '.cc', '.cxx', '.cppm', '.c++', '.cp', '.C', '.CPP',
                  '.hpp', '.hh', '.hxx',
                  '.ii', '.iim')
llvm_extensions = ('.bc', '.ll')


def create_working_directory(wd=None, save=False):
    ''' Create a temporary working directory '''
    if not wd:
        wd = tempfile.mkdtemp(prefix='ikos-')

    if not os.path.exists(wd):
        try:
            os.makedirs(wd)
        except OSError as e:
            printf('error: %s: %s\n', wd, e.strerror, file=sys.stderr)
            sys.exit(1)

    if not os.path.isdir(wd):
        printf('error: %s: Not a directory\n', wd, file=sys.stderr)
        sys.exit(1)

    if not save:
        atexit.register(shutil.rmtree, path=wd)
    else:
        log.info('Temporary files will be kept in directory: %s' % wd)

    return wd


def namer(path, ext, wd):
    '''
    Return the path to a file with the given extension,
    in the given working directory.

    >>> namer('/home/me/test.c', '.bc', '/tmp/ikos-xxx')
    '/tmp/ikos-xxx/test.bc'
    '''
    base = os.path.basename(path)
    return os.path.join(wd, os.path.splitext(base)[0] + ext)


def signal_name(signum):
    ''' Return the signal name given the signal number '''
    for name, value in signal.__dict__.items():
        if (name.startswith('SIG') and
                not name.startswith('SIG_') and
                value == signum):
            return name
    return str(signum)


def is_apron_domain(domain):
    ''' Return True if the given domain is an APRON numerical domain '''
    return 'apron-' in domain


domains_without_narrowing = (
    'apron-polka-polyhedra',
    'apron-polka-linear-equalities',
    'apron-ppl-polyhedra',
    'apron-ppl-linear-congruences',
    'apron-pkgrid-polyhedra-lin-cong',
    'var-pack-apron-polka-polyhedra',
    'var-pack-apron-polka-linear-equalities',
    'var-pack-apron-ppl-polyhedra',
    'var-pack-apron-ppl-linear-congruences',
    'var-pack-apron-pkgrid-polyhedra-lin-cong',
)


def clang_emit_llvm_flags():
    ''' Clang flags to emit llvm bitcode '''
    return ['-c', '-emit-llvm']


def clang_ikos_flags():
    ''' Clang flags for ikos '''
    return [
        # enable clang warnings
        '-Wall',
        # disable source code fortification
        '-U_FORTIFY_SOURCE',
        '-D_FORTIFY_SOURCE=0',
        # flag for intrinsic.h
        '-D__IKOS__',
        # compile in debug mode
        '-g',
        # disable optimizations
        '-O0',
        # disable the 'optnone' attribute
        # see https://bugs.llvm.org/show_bug.cgi?id=35950#c10
        '-Xclang',
        '-disable-O0-optnone',
    ]


##################
# ikos toolchain #
##################

def clang(
    bc_path,
    cpp_path,
    include_flags=None,
    define_flags=None,
    warning_flags=None,
    disable_warnings=False,
    machine_flags=None,
    colors=True,
):
    cmd = [settings.clang()]
    cmd += clang_emit_llvm_flags()
    cmd += clang_ikos_flags()
    cmd += [cpp_path,
            '-o',
            bc_path]

    # For #include <ikos/analyzer/intrinsic.hpp>
    cmd += ['-isystem', settings.INCLUDE_DIR]

    if include_flags:
        cmd += ['-I%s' % i for i in include_flags]
    if define_flags:
        cmd += ['-D%s' % d for d in define_flags]
    if warning_flags:
        cmd += ['-W%s' % w for w in warning_flags]
    if disable_warnings:
        cmd.append('-w')
    if machine_flags:
        cmd += ['-m%s' % m for m in machine_flags]

    if colors:
        cmd.append('-fcolor-diagnostics')
    else:
        cmd.append('-fno-color-diagnostics')

    if path_ext(cpp_path) in cpp_extensions:
        cmd.append('-std=c++17')  # available because clang >= 7.0

    log.info('Compiling %s' % cpp_path)
    log.debug('Running %s' % command_string(cmd))
    subprocess.check_call(cmd)


def ikos_pp(pp_path, bc_path, entry_points, opt_level, inline_all, verify):
    if opt_level == 'aggressive':
        log.warning('Using aggressive optimizations is not recommended')
        log.warning('The translation from LLVM bitcode to AR might fail')

    cmd = [settings.ikos_pp(),
           '-opt=%s' % opt_level,
           '-entry-points=%s' % ','.join(entry_points)]

    if inline_all:
        cmd.append('-inline-all')

    if not verify:
        cmd.append('-no-verify')

    cmd += [bc_path, '-o', pp_path]

    log.info('Running ikos preprocessor')
    log.debug('Running %s' % command_string(cmd))
    subprocess.check_call(cmd)


def display_llvm(pp_path):
    log.info('Printing LLVM')
    cmd = [settings.opt(), '-S', pp_path]
    subprocess.check_call(cmd)


class AnalyzerError(Exception):
    def __init__(self, message, cmd, returncode):
        super(AnalyzerError, self).__init__(message)
        self.cmd = cmd
        self.returncode = returncode


def ikos_analyzer(db_path, pp_path, opt):
    if settings.BUILD_MODE == 'Debug':
        log.warning('ikos was built in debug mode, the analysis might be slow')
    if is_apron_domain(opt.domain) and opt.jobs != 1:
        log.warning('apron abstract domains are not thread-safe, '
                    'the analysis might crash')

    # Fix huge slow down when ikos-analyzer uses DROP TABLE on an existing db
    if os.path.isfile(db_path):
        os.remove(db_path)

    cmd = [settings.ikos_analyzer()]

    # analysis options
    cmd += ['-a=%s' % ','.join(opt.analyses),
            '-d=%s' % opt.domain,
            '-entry-points=%s' % ','.join(opt.entry_points),
            '-globals-init=%s' % opt.globals_init,
            '-proc=%s' % opt.procedural,
            '-j=%d' % opt.jobs,
            '-widening-strategy=%s' % opt.widening_strategy,
            '-widening-delay=%d' % opt.widening_delay,
            '-widening-period=%d' % opt.widening_period]

    if opt.narrowing_strategy == 'auto':
        if opt.domain in domains_without_narrowing:
            cmd.append('-narrowing-strategy=meet')
        else:
            cmd.append('-narrowing-strategy=narrow')
    else:
        cmd.append('-narrowing-strategy=%s' % opt.narrowing_strategy)

    if opt.narrowing_iterations is not None:
        cmd.append('-narrowing-iterations=%d' % opt.narrowing_iterations)
    elif (opt.narrowing_strategy == 'auto' and
            opt.domain in domains_without_narrowing):
        cmd.append('-narrowing-iterations=%d'
                   % args.meet_iterations_if_no_narrowing)

    if opt.widening_delay_functions:
        cmd.append('-widening-delay-functions=%s'
                   % ','.join(opt.widening_delay_functions))

    if opt.no_init_globals:
        cmd.append('-no-init-globals=%s' % ','.join(opt.no_init_globals))
    if opt.no_liveness:
        cmd.append('-no-liveness')
    if opt.no_pointer:
        cmd.append('-no-pointer')
    if opt.no_widening_hints:
        cmd.append('-no-widening-hints')
    if opt.partitioning != 'no':
        cmd.append('-enable-partitioning-domain')
    if opt.no_fixpoint_cache:
        cmd.append('-no-fixpoint-cache')
    if opt.no_checks:
        cmd.append('-no-checks')
    if opt.hardware_addresses:
        cmd.append('-hardware-addresses=%s' % ','.join(opt.hardware_addresses))
    if opt.hardware_addresses_file:
        cmd.append('-hardware-addresses-file=%s' % opt.hardware_addresses_file)
    if opt.argc is not None:
        cmd.append('-argc=%d' % opt.argc)

    # import options
    cmd.append('-allow-dbg-mismatch')
    if opt.no_bc_verify:
        cmd.append('-no-verify')
    if opt.no_libc:
        cmd.append('-no-libc')
    if opt.no_libcpp:
        cmd.append('-no-libcpp')
    if opt.no_libikos:
        cmd.append('-no-libikos')

    # AR passes options
    if opt.no_type_check:
        cmd.append('-no-type-check')
    if opt.no_simplify_cfg:
        cmd.append('-no-simplify-cfg')
    if opt.no_simplify_upcast_comparison:
        cmd.append('-no-simplify-upcast-comparison')
    if 'gauge' in opt.domain:
        cmd.append('-add-loop-counters')
    if opt.partitioning == 'return':
        cmd.append('-add-partitioning-variables')

    # debug options
    cmd += ['-display-checks=%s' % opt.display_checks,
            '-display-inv=%s' % opt.display_inv]

    if opt.display_ar:
        cmd.append('-display-ar')
    if opt.trace_ar_stmts:
        cmd.append('-trace-ar-stmts')
    if opt.display_liveness:
        cmd.append('-display-liveness')
    if opt.display_function_pointer:
        cmd.append('-display-function-pointer')
    if opt.display_pointer:
        cmd.append('-display-pointer')
    if opt.display_fixpoint_parameters:
        cmd.append('-display-fixpoint-parameters')
    if opt.generate_dot:
        cmd += ['-generate-dot', '-generate-dot-dir', opt.generate_dot_dir]

    # add -name-values if necessary
    if (opt.display_checks in ('all', 'fail') or
            opt.display_inv in ('all', 'fail') or
            opt.display_liveness or
            opt.display_fixpoint_parameters or
            opt.display_function_pointer or
            opt.display_pointer or
            opt.display_raw_checks):
        cmd.append('-name-values')

    # misc. options
    if opt.color == 'yes':
        cmd.append('-color=1')
    elif opt.color == 'no':
        cmd.append('-color=0')

    cmd.append('-log=%s' % opt.log_level)
    cmd.append('-progress=%s' % opt.progress)

    # input/output
    cmd += [pp_path, '-o', db_path]

    # set resource limit, if requested
    if opt.mem:
        import resource  # fails on Windows

        def set_limits():
            mem_bytes = opt.mem * 1024 * 1024
            resource.setrlimit(resource.RLIMIT_AS, [mem_bytes, mem_bytes])
    else:
        set_limits = None

    # called after timeout
    def kill(p):
        try:
            log.error('Timeout')
            p.send_signal(signal.SIGALRM)
        except OSError:
            pass

    log.info('Running ikos analyzer')
    log.debug('Running %s' % command_string(cmd))
    p = subprocess.Popen(cmd, preexec_fn=set_limits)
    timer = threading.Timer(opt.cpu, kill, [p])

    if opt.cpu:
        timer.start()

    try:
        if sys.platform.startswith('win'):
            return_status = p.wait()
        else:
            _, return_status = os.waitpid(p.pid, 0)
    finally:
        # kill the timer if the process has terminated already
        if timer.is_alive():
            timer.cancel()

    # special case for Windows, since it does not define WIFEXITED & co.
    if sys.platform.startswith('win'):
        if return_status != 0:
            raise AnalyzerError('a run-time error occurred', cmd, return_status)
        else:
            return

    # if it did not terminate properly, propagate this error code
    if os.WIFEXITED(return_status) and os.WEXITSTATUS(return_status) != 0:
        exit_status = os.WEXITSTATUS(return_status)
        raise AnalyzerError('a run-time error occurred', cmd, exit_status)

    if os.WIFSIGNALED(return_status):
        signum = os.WTERMSIG(return_status)
        raise AnalyzerError('exited with signal %s' % signal_name(signum),
                            cmd,
                            signum)

    if os.WIFSTOPPED(return_status):
        signum = os.WSTOPSIG(return_status)
        raise AnalyzerError('exited with signal %d' % signal_name(signum),
                            cmd,
                            signum)


def ikos_view(opt, db):
    from ikos import view
    v = view.View(db)
    v.serve()


#################
# main for ikos #
#################

def main(argv):
    progname = os.path.basename(argv[0])

    start_date = datetime.datetime.now()

    # parse arguments
    opt = parse_arguments(argv[1:])

    # setup colors and logging
    colors.setup(opt.color, file=log.out)
    log.setup(opt.log_level)

    if is_apron_domain(opt.domain) and not settings.HAS_APRON:
        printf('%s: error: cannot use apron abstract domains.\n'
               'ikos was compiled without apron support, '
               'see analyzer/README.md\n',
               progname, file=sys.stderr)
        sys.exit(1)

    # create working directory
    wd = create_working_directory(opt.temp_dir, opt.save_temps)

    input_path = opt.file

    # compile c/c++ code
    if path_ext(input_path) in c_extensions + cpp_extensions:
        bc_path = namer(opt.file, '.bc', wd)

        try:
            with stats.timer('clang'):
                clang(bc_path, input_path,
                      opt.compiler_include_flags,
                      opt.compiler_define_flags,
                      opt.compiler_warning_flags,
                      opt.compiler_disable_warnings,
                      opt.compiler_machine_flags,
                      colors.ENABLE)
        except subprocess.CalledProcessError as e:
            printf('%s: error while compiling %s, abort.\n',
                   progname, input_path, file=sys.stderr)
            sys.exit(e.returncode)

        input_path = bc_path

    if path_ext(input_path) not in llvm_extensions:
        printf('%s: error: unexpected file extension.\n',
               progname, file=sys.stderr)
        sys.exit(1)

    # ikos-pp: preprocess llvm bitcode
    pp_path = namer(opt.file, '.pp.bc', wd)
    try:
        with stats.timer('ikos-pp'):
            ikos_pp(pp_path, input_path,
                    opt.entry_points, opt.opt_level,
                    opt.inline_all, not opt.no_bc_verify)
    except subprocess.CalledProcessError as e:
        printf('%s: error while preprocessing llvm bitcode, abort.\n',
               progname, file=sys.stderr)
        sys.exit(e.returncode)

    # display the llvm bitcode, if requested
    if opt.display_llvm:
        display_llvm(pp_path)

    # ikos-analyzer: analyze llvm bitcode
    try:
        with stats.timer('ikos-analyzer'):
            ikos_analyzer(opt.output_db, pp_path, opt)
    except AnalyzerError as e:
        printf('%s: error: %s\n', progname, e, file=sys.stderr)
        sys.exit(e.returncode)

    # open output database
    db = OutputDatabase(path=opt.output_db)

    # insert timing results in the database
    db.insert_timing_results(stats.rows())

    # insert settings in the database
    settings_rows = [
        ('version', settings.VERSION),
        ('start-date', start_date.isoformat(' ')),
        ('end-date', datetime.datetime.now().isoformat(' ')),
        ('working-directory', wd),
        ('input', opt.file),
        ('bc-file', input_path),
        ('pp-bc-file', pp_path),
        ('clang', settings.clang()),
        ('ikos-pp', settings.ikos_pp()),
        ('opt-level', opt.opt_level),
        ('inline-all', json.dumps(opt.inline_all)),
        ('use-libc-intrinsics', json.dumps(not opt.no_libc)),
        ('use-libcpp-intrinsics', json.dumps(not opt.no_libcpp)),
        ('use-libikos-intrinsics', json.dumps(not opt.no_libikos)),
        ('use-simplify-cfg', json.dumps(not opt.no_simplify_cfg)),
        ('use-simplify-upcast-comparison',
         json.dumps(not opt.no_simplify_upcast_comparison)),
    ]
    if opt.cpu:
        settings_rows.append(('cpu-limit', opt.cpu))
    if opt.mem:
        settings_rows.append(('mem-limit', opt.mem))
    db.insert_settings(settings_rows)

    first = (log.LEVEL >= log.ERROR)

    # display timing results
    if opt.display_times != 'no':
        if not first:
            printf('\n')
        report.print_timing_results(db, opt.display_times == 'full')
        first = False

    # display summary
    if opt.display_summary != 'no':
        if not first:
            printf('\n')
        report.print_summary(db, opt.display_summary == 'full')
        first = False

    # display raw checks
    if opt.display_raw_checks:
        if not first:
            printf('\n')
        report.print_raw_checks(db, opt.procedural == 'inter')
        first = False

    # start ikos-view
    if opt.format == 'web':
        ikos_view(opt, db)
        return

    # report
    if opt.format != 'no':
        if not first and opt.report_file is sys.stdout:
            printf('\n' + colors.bold('# Results') + '\n')
            first = False

        # setup colors again (in case opt.color = 'auto')
        colors.setup(opt.color, file=opt.report_file)

        # generate report
        rep = report.generate_report(db,
                                     status_filter=opt.status_filter,
                                     analyses_filter=None)

        # format report
        formatter_class = report.formats[opt.format]
        formatter = formatter_class(opt.report_file, opt.report_verbosity)
        formatter.format(rep)

    # close database
    db.close()

    if opt.remove_db:
        os.remove(opt.output_db)



###############################################################################
#
# Color utilities
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2011-2019 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import os
import sys


# are colors enable?
ENABLE = True

# available colors
COLORS = {
    'grey': 0,
    'red': 1,
    'green': 2,
    'yellow': 3,
    'blue': 4,
    'magenta': 5,
    'cyan': 6,
    'white': 7,
}

# available attributes
ATTRIBUTES = {
    'bold': 1,
    'dark': 2,
    'underline': 4,
    'blink': 5,
    'reverse': 7,
    'concealed': 8
}


def setup(color, file=sys.stdout):
    global ENABLE

    if color in ('yes', 'on', 'always'):
        ENABLE = True
    elif color in ('no', 'off', 'never'):
        ENABLE = False
    elif color == 'auto':
        ENABLE = os.isatty(file.fileno())


def colorize(text, color=None, on_color=None, attrs=None):
    '''
    Colorize text.

    >>> colorize('Hello, World!', 'red', 'grey', ['bold', 'blink'])
    '\x1b[5m\x1b[1m\x1b[40m\x1b[31mHello, World!\x1b[0m'
    >>> colorize('Hello, World!', 'green')
    '\x1b[32mHello, World!\x1b[0m'
    '''
    if ENABLE:
        if color:
            text = '\033[%dm%s' % (30 + COLORS[color], text)

        if on_color:
            text = '\033[%dm%s' % (40 + COLORS[on_color], text)

        if attrs:
            for attr in attrs:
                text = '\033[%dm%s' % (ATTRIBUTES[attr], text)

        text += '\033[0m'

    return text


###########
# helpers #
###########


def bold(s):
    return colorize(s, attrs=['bold'])


for color in COLORS:
    globals()[color] = (lambda s, color=color: colorize(s, color))
    globals()['bold_%s' % color] = (lambda s, color=color:
                                    colorize(s, color, attrs=['bold']))
    globals()['on_%s' % color] = (lambda s, color=color:
                                  colorize(s, on_color=color))



###############################################################################
#
# Simple statistics module
#
# Author: Jorge A. Navas
#
# Contributors: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2011-2019 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import time


# use monotonic clock if available
if hasattr(time, 'monotonic'):
    _current_time = time.monotonic
else:
    _current_time = time.time


class Stopwatch:
    ''' A stop watch '''

    def __init__(self):
        self._started = 0
        self._finished = -1
        self._elapsed = 0
        self.start()

    @property
    def elapsed(self):
        ''' Returns time (in seconds) since the stopwatch has been started '''
        if self._finished < self._started:
            return self._elapsed + (_current_time() - self._started)
        else:
            return self._elapsed + (self._finished - self._started)

    def start(self):
        ''' Starts or resumes the stopwatch '''
        # collect elapsed time so far
        if self._finished >= self._started:
            self._elapsed += (self._finished - self._started)

        self._started = _current_time()
        self._finished = -1

    def stop(self):
        ''' Stops the stopwatch '''
        if self._finished < self._started:
            self._finished = _current_time()

    def reset(self):
        ''' Resets the stopwatch by erasing all elapsed time '''
        self._elapsed = 0
        self._finished = -1
        self.start()

    def __str__(self):
        ''' Reports time in seconds up to three decimal places '''
        return '{0:.3f}'.format(self.elapsed)


_statistics = dict()


def get(key):
    ''' Gets a value from statistics table '''
    return _statistics.get(key)


def put(key, v):
    ''' Puts a value in statistics table '''
    _statistics[key] = v


def start(key):
    ''' Starts (or resumes) a named stopwatch '''
    sw = get(key)
    if sw is None:
        sw = Stopwatch()
        put(key, sw)
        return sw
    else:
        sw.start()


def stop(key):
    ''' Stops a named stopwatch '''
    sw = get(key)
    if sw is not None:
        sw.stop()


def rows():
    ''' Return all statistics as rows '''
    return [(name, sw.elapsed) for name, sw in _statistics.items()]


def timer(key):
    '''
    ContextManager to help measuring time.

    with timer('myname') as t:
        do_code_that_is_timed
    '''

    class TimerContextManager(object):
        def __init__(self, key):
            self._key = key

        def __enter__(self):
            start(self._key)
            return None

        def __exit__(self, exc_type, exc_value, traceback):
            stop(self._key)
            return False

    return TimerContextManager(key)



###############################################################################
#
# Highlight utility for ikos-view
#
# Author: Thomas Bailleux
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2018-2023 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import html

from ikos import log


class DummyLexer(object):
    def __init__(self, *args, **kwargs):
        pass


class DummyHtmlFormatter(object):
    def __init__(self):
        pass

    def get_style_defs(self, cl):
        return ''

    def wrap(self, src):
        pass


def dummy_highlight(data, lexer, formatter):
    lines = ((1, html.escape(line)) for line in data.split('\n'))
    code = ''
    for _, line in formatter.wrap(lines):
        code += line
    return code


try:
    from pygments.lexers import CLexer, CppLexer
    from pygments.formatters import HtmlFormatter
    from pygments import highlight
except ImportError:
    log.warning("Syntax highlighting is not available:"
                " module `pygments` is not installed.")
    log.warning("Note: use `pip install --user pygments`"
                " to install `pygments`")
    CLexer = DummyLexer
    CppLexer = DummyLexer
    HtmlFormatter = DummyHtmlFormatter
    highlight = dummy_highlight



###############################################################################
#
# Wrapper for python standard library HTTP server
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2018-2023 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################

import os
from errno import EAFNOSUPPORT
from http.server import HTTPServer, BaseHTTPRequestHandler
from socket import AF_INET, AF_INET6
from urllib.parse import parse_qs, urlencode
from urllib.request import urlopen

class HTTPServerIPv6(HTTPServer):
    '''
    HTTP server that listens on IPv6 when available.

    In systems with dual stack, this also listens on IPv4.

    This class tries to listen on IPv6 first (and IPv4 if available). If that
    is not supported, then it tries IPv4 only.
    '''

    address_family = AF_INET6

    def __init__(self, server_address, RequestHandlerClass, bind_and_activate=True):
        try:
            # Initially try to open the server with IPv6. This will also select
            # IPv4 on systems with dual stack.
            super().__init__(server_address, RequestHandlerClass, bind_and_activate)

        except OSError as e:
            if e.errno == EAFNOSUPPORT:
                # If the exception is due to IPv6 not being supported, we
                # select IPv4 only and try to re-open the server again.
                self.address_family = AF_INET
                super().__init__(server_address, RequestHandlerClass, bind_and_activate)
            else:
                # If the exception is for any other reason other than IPv6 not
                # being supported, then we cannot do anything about it.
                raise e



###############################################################################
#
# File type detection utility
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2018-2019 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################
import subprocess
import sys

from ikos.log import printf


def filetype(path):
    ''' Return the file type as returned by the 'file' command '''
    try:
        output = subprocess.check_output(['file', '--brief', path])
    except OSError as e:
        printf('error: file: %s', e.strerror, file=sys.stderr)
        sys.exit(e.errno)

    if output.startswith(b'cannot open '):
        open(path)  # raise IOError

    return output.decode('utf-8').strip()



###############################################################################
#
# Abstract interpretation utilities for python
#
# Author: Maxime Arthaud
#
# Contact: ikos@lists.nasa.gov
#
# Notices:
#
# Copyright (c) 2018-2019 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Disclaimers:
#
# No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
# ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
# TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
# ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
# OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
# ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
# THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
# ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
# RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
# RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
# DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
# IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
#
# Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
# THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
# AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
# IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
# USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
# RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
# HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
# AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
# RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
# UNILATERAL TERMINATION OF THIS AGREEMENT.
#
###############################################################################


class Signedness:
    SIGNED = 0
    UNSIGNED = 1


def assert_compatible(a, b):
    ''' Check that the given parameters have the same bit-width and sign '''
    assert a.bit_width == b.bit_width and a.sign == b.sign


class MachineInt:
    ''' Represents a machine integer '''

    __slots__ = ('n', 'bit_width', 'sign')

    def __init__(self, n, bit_width, sign):
        self.n = n
        self.bit_width = bit_width
        self.sign = sign
        self._normalize()

    def _normalize(self):
        if self.is_signed():
            # n needs to be within [-2**(n-1), 2**(n-1)-1]
            self.n = self.n + 2**(self.bit_width - 1)
            self.n = self.n % 2**self.bit_width
            self.n = self.n - 2**(self.bit_width - 1)
        else:
            # n needs to be within [0, 2**n-1]
            self.n = self.n % 2**self.bit_width

    def is_signed(self):
        return self.sign == Signedness.SIGNED

    def is_unsigned(self):
        return self.sign == Signedness.UNSIGNED

    @staticmethod
    def min(bit_width, sign):
        if sign == Signedness.SIGNED:
            return MachineInt(-2**(bit_width - 1), bit_width, sign)
        else:
            return MachineInt(0, bit_width, sign)

    def is_min(self):
        if self.is_signed():
            return self.n == -2**(self.bit_width - 1)
        else:
            return self.n == 0

    @staticmethod
    def max(bit_width, sign):
        if sign == Signedness.SIGNED:
            return MachineInt(2**(bit_width - 1) - 1, bit_width, sign)
        else:
            return MachineInt(2**bit_width - 1, bit_width, sign)

    def is_max(self):
        if self.is_signed():
            return self.n == 2**(self.bit_width - 1) - 1
        else:
            return self.n == 2**self.bit_width - 1

    def high_bit(self):
        if self.is_signed():
            return self.n < 0
        else:
            return self.n >= 2**(self.bit_width - 1)

    def sign_cast(self, sign):
        return MachineInt(self.n, self.bit_width, sign)

    def __eq__(self, other):
        assert_compatible(self, other)
        return self.n == other.n

    def __lt__(self, other):
        assert_compatible(self, other)
        return self.n < other.n

    def __le__(self, other):
        assert_compatible(self, other)
        return self.n <= other.n

    def __gt__(self, other):
        assert_compatible(self, other)
        return self.n > other.n

    def __ge__(self, other):
        assert_compatible(self, other)
        return self.n >= other.n

    def __str__(self):
        return str(self.n)


class Interval:
    ''' Represents an interval [a, b], or bottom '''

    __slots__ = ('lb', 'ub')

    def __init__(self, lb, ub):
        assert_compatible(lb, ub)
        self.lb = lb
        self.ub = ub

    @staticmethod
    def top(bit_width, sign):
        return Interval(MachineInt.min(bit_width, sign),
                        MachineInt.max(bit_width, sign))

    @staticmethod
    def bottom(bit_width, sign):
        return Interval(MachineInt.max(bit_width, sign),
                        MachineInt.min(bit_width, sign))

    @staticmethod
    def from_dict(info):
        ''' Create an interval from a dictionary '''
        t = info['type']
        sign = Signedness.SIGNED if t[0] == 's' else Signedness.UNSIGNED
        bit_width = int(t[1:])
        return Interval(MachineInt(info['lb'], bit_width, sign),
                        MachineInt(info['ub'], bit_width, sign))

    @property
    def bit_width(self):
        return self.lb.bit_width

    @property
    def sign(self):
        return self.lb.sign

    def is_bottom(self):
        return self.lb > self.ub

    def is_top(self):
        return self.lb.is_min() and self.ub.is_max()

    def is_constant(self):
        return self.lb == self.ub

    def to_constraints(self, var):
        assert not self.is_bottom()

        if self.is_top():
            return 'could not bound %s' % var
        elif self.lb == self.ub:
            return '%s = %s' % (var, self.lb)
        elif self.lb.is_min():
            return '%s <= %s' % (var, self.ub)
        elif self.ub.is_max():
            return '%s >= %s' % (var, self.lb)
        else:
            return '%s <= %s <= %s' % (self.lb, var, self.ub)

    def sign_cast(self, sign):
        if self.sign == sign:
            return self
        elif self.is_bottom():
            return Interval.bottom(self.bit_width, sign)
        else:
            if self.lb.high_bit() == self.ub.high_bit():
                lb = self.lb.sign_cast(sign)
                ub = self.ub.sign_cast(sign)
                if lb <= ub:
                    return Interval(lb, ub)
                else:
                    return Interval(ub, lb)
            else:
                return Interval.top(self.bit_width, sign)

    def __str__(self):
        if self.is_bottom():
            return '⊥'
        elif self.is_top():
            return 'T'
        else:
            return '[%s, %s]' % (self.lb, self.ub)


class Congruence:
    ''' Represents a congruence aZ + b, or bottom '''

    __slots__ = ('a', 'b', 'bit_width', 'sign', 'bottom')

    def __init__(self, a, b, bit_width, sign, bottom):
        self.a = a
        self.b = b
        self.bit_width = bit_width
        self.sign = sign
        self.bottom = bottom
        self._normalize()

    def _normalize(self):
        if self.bottom:
            self.a = self.b = None
            return

        assert self.a >= 0

        if self.a != 0:
            self.b = self.b % self.a

    @staticmethod
    def from_dict(info):
        ''' Create a congruence from a dictionary '''
        t = info['type']
        sign = Signedness.SIGNED if t[0] == 's' else Signedness.UNSIGNED
        bit_width = int(t[1:])
        return Congruence(a=info['a'],
                          b=info['b'],
                          bit_width=bit_width,
                          sign=sign,
                          bottom=False)

    def is_bottom(self):
        return self.bottom

    def is_top(self):
        return not self.bottom and self.a == 1

    def is_constant(self):
        return not self.bottom and self.a == 0

    def __str__(self):
        if self.is_bottom():
            return '⊥'
        if self.a == 0:
            return str(self.b)
        if self.b == 0:
            return '%dZ' % self.a
        return '%dZ+%d' % (self.a, self.b)
